"""
Apex Legends API

| The ApexLegendsAPI wraps the api at: https://apexlegendsapi.com
| Get your API Key Here: https://apexlegendsapi.com
"""
import json
import requests
from deprecated import deprecated
from .al_domain import ALPlayer  # noqa E0402
from .al_base import ALPlatform, ALAction, ALHTTPExceptionFromResponse  # noqa E0402


class ApexLegendsAPI:
    """
    Main class that wraps the API calls
    """
    api_version: str = "5"
    base_params: dict = {'version': api_version}
    base_url: str = "https://api.mozambiquehe.re/bridge"

    def __init__(self, api_key: str):
        """ Initialize with the API Key """
        self.session: requests.Session = requests.Session()
        self.session.headers.update({'Authorization': api_key})

    def _make_request(self, additional_params: dict, new_base_url: str = None) -> list:
        """ Send the request to the apex legends api """
        if new_base_url:
            url: str = new_base_url
            params: dict = additional_params
        else:
            url: str = self.base_url
            params: dict = dict(self.base_params, **additional_params)
        response: requests.Response = self.session.get(url, params=params)
        if response.status_code == 200:
            try:
                response_text = json.loads(response.text)
            except ValueError:
                response_text = response.text
        else:
            raise ALHTTPExceptionFromResponse(response)

        # sometimes we get a pure dictionary back, let's wrap it in a list for consistency
        if isinstance(response_text, dict):
            response_text = [response_text]
        return response_text

    def nametouid(self, player: str, platform: ALPlatform) -> int:
        """
        Retrieve a player's uid given they're name and platform

        :parameter player: Name of the player
        :type player: str
        :parameter platform: see [ALPlatform] for all types
        :type platform: ALPlatform
        """

        new_base_url: str = "https://api.mozambiquehe.re/nametouid?"
        additional_params = {
            'player': player,
            'platform': platform.value
        }
        result = self._make_request(additional_params=additional_params, new_base_url=new_base_url)
        assert len(result) == 1
        return result[0]['result']

    def get_player(self, name: str, platform: ALPlatform, skip_tracker_rank=False) -> ALPlayer:
        """
        Retrieve the ALPlayer object populated with data from the api.

        NOTE:
            Player must exist, method will return None if the player cannot be found

        :parameter name: Name of the player
        :type name: str
        :parameter platform: see ALPlatform for all types
        :parameter skip_tracker_rank: if set to True, this will skip fetching the legend ranks
        :return: a single player or None if no player is found
        """
        basic_player_stats: list = self.basic_player_stats(name, platform, skip_tracker_rank)
        assert len(basic_player_stats) == 1
        event_info: list = self.events(
            player_name=name,
            platform=platform,
            action=ALAction.INFO
        )
        events: list = list()
        tracked_player: dict
        for tracked_player in event_info[0].get('data'):
            if name == tracked_player.get('name') and \
                    platform.value == tracked_player.get('platform'):
                events = self.events(
                    player_name=name,
                    platform=platform,
                    action=ALAction.GET
                )
        return ALPlayer(basic_player_stats_data=basic_player_stats[0], events=events)

    def get_player_by_uid(
            self, uid: str, platform: ALPlatform, skip_tracker_rank=False
    ) -> ALPlayer:
        """
        Retrieve the ALPlayer object populated with data from the api.

        NOTE:
            Player must exist, method will return None if the player cannot be found

        :parameter uid: UID of the player
        :parameter platform: see ALPlatform for all types
        :parameter skip_tracker_rank: if set to True, this will skip fetching the legend ranks
        :return: a single player or None if no player is found
        """
        basic_player_stats: list = self.basic_player_stats_by_uid(uid, platform, skip_tracker_rank)
        assert len(basic_player_stats) == 1
        event_info: list = self.events_by_uid(
            uid=uid,
            platform=platform,
            action=ALAction.INFO
        )
        events: list = list()
        tracked_player: dict
        for tracked_player in event_info[0].get('data'):
            if uid == tracked_player.get('uid') and \
                    platform.value == tracked_player.get('platform'):
                events = self.events_by_uid(
                    uid=uid,
                    platform=platform,
                    action=ALAction.GET
                )
        return ALPlayer(basic_player_stats_data=basic_player_stats[0], events=events)

    def add_player_by_uid(self, player_uid: int, platform: ALPlatform) -> list:
        """
        Adds the given player's UUID to the list of tracked players

        :param player_uid: UUID of the player to add
        :type player_uid: int
        :param platform: ALPlatform of player to add
        :type platform: ALPlatform
        """
        return self.events_by_uid(str(player_uid), platform=platform, action=ALAction.ADD)

    def basic_player_stats(
            self, player_name: str,
            platform: ALPlatform,
            skip_tracker_rank=False) -> list:
        """
        Query the server for the given player / platform and returns a dictionary of their
        stats.
        More here: https://apexlegendsapi.com/#basic

        :param player_name: Player Name to search for
        :param platform: (see Platform enum for values)
        :param skip_tracker_rank: if set to true, this will not fetch the legend's tracker rank
        :return: List of player stats created from response json
        """
        params: dict = {'platform': platform.value, 'player': player_name}
        if skip_tracker_rank:
            params.update({'skipRank': True})
        return self._make_request(additional_params=params)

    def basic_player_stats_by_uid(
            self, uid: str,
            platform: ALPlatform,
            skip_tracker_rank=False) -> list:
        """
        Query the server for the given player / platform and returns a dictionary of their
        stats.
        More here: https://apexlegendsapi.com/#basic

        :param uid: Player UID to search for
        :param platform: (see Platform enum for values)
        :param skip_tracker_rank: if set to true, this will not fetch the legend's tracker rank
        :return: List of player stats created from response json
        """
        params: dict = {'platform': platform.value, 'uid': uid}
        if skip_tracker_rank:
            params.update({'skipRank': True})
        return self._make_request(additional_params=params)

    @deprecated(reason="use `events` instead")
    def match_history(self, player_name: str, platform: ALPlatform, action: ALAction) -> list:
        """
        .. deprecated:: 1.1.0
            use `events` instead
        """
        return self.events(player_name=player_name, platform=platform, action=action)

    def events(self, player_name: str, platform: ALPlatform, action: ALAction) -> list:
        """
        Query the server for the given player / platform and return a list of their
        events

        NOTE:
          * Match history is only available for supporters
          * Match history must be tracked by the server otherwise this will return nothing
          * In order to add a player to be tracked, you need to call this passing 'add' action.

        :param player_name: Player Name for match history
        :param platform: see Platform enum for values
        :param action: see Action enum for values
        :return: List of history created from response json
        """
        params: dict = {
            'platform': platform.value,
            'player': player_name,
            'history': 1,
            'action': action.value
        }
        return self._make_request(additional_params=params)

    def events_by_uid(self, uid: str, platform: ALPlatform, action: ALAction) -> list:
        """
        Query the server for the given player's UID / platform and return a list of their
        events

        NOTE:
          * Match history is only available for supporters
          * Match history must be tracked by the server otherwise this will return nothing
          * In order to add a player to be tracked, you need to call this passing 'add' action.

        :param uid: Player UID for match history
        :param platform: see Platform enum for values
        :param action: see Action enum for values
        :return: List of history created from response json
        """
        params: dict = {
            'platform': platform.value,
            'uid': uid,
            'history': 1,
            'action': action.value
        }
        return self._make_request(additional_params=params)

    def get_player_origin(self, player_name: str, show_all_hits: bool = False) -> list:
        """
        Query the server for the origin user and returns Origin UID, real username, PID and avatar
         for a given username

        :param player_name: Player Name for match history
        :param show_all_hits: True to 'search' for player (show multiple hits), default False
        :return: list of results
        """
        new_base_url: str = "https://api.mozambiquehe.re/origin?"
        new_base_url += f"&player={player_name}"
        if show_all_hits:
            new_base_url += "&showAllHits"
        return self._make_request(additional_params={}, new_base_url=new_base_url)



"""
Player class for the Apex Legends API Python package
"""
from typing import List
import arrow
from .al_base import ALEventType, ALPlatform  # noqa E0402


# pylint: disable=too-few-public-methods
class GameInfo:
    """
    Data structure for game information for selected legend

    Example:
        - skin=Extreme Measures
        - frame=Fuel Injected
        - pose=Spin and Flick
        - intro=Run fast, hit fast, win fast
    """
    class Badge:
        """ data structure for badges """
        def __init__(self, badge_dict: dict):
            self.name: str = badge_dict.get('name')
            """ Name of the badge """
            self.value: str = badge_dict.get('value')
            """ Value of the badge """
            if badge_dict.get('category'):
                self.category: str = badge_dict.get('category')
                """
                Category of the badge

                Note:
                    - This is either 'Account Badges' for account level badge, or the legend name
                """

    def __init__(self, game_info_dict: dict):
        if game_info_dict.get('skin'):
            self.skin: str = game_info_dict.get('skin')
            """ Skin on the current legend """
        if game_info_dict.get('frame'):
            self.frame: str = game_info_dict.get('frame')
            """ Frame chosen for the current legend """
        if game_info_dict.get('pose'):
            self.pose: str = game_info_dict.get('pose')
            """ Pose of the current legend """
        if game_info_dict.get('intro'):
            self.intro: str = game_info_dict.get('intro')
            """ Intro quip of the current legend """
        self.badges: list[GameInfo.Badge] = list()
        """ Currently selected 'badges' for the current legend """
        for badge in game_info_dict.get('badges'):
            self.badges.append(GameInfo.Badge(badge))


class Event:
    """ Parent class for apex-legend events """
    def __init__(self, event_dict: dict):
        self.uid: int = event_dict.get('uid')
        """ Origin / EA UUID of the player """
        self.player: str = event_dict.get('player')
        """ Player nick name """
        self.timestamp: int = event_dict.get('timestamp')
        """ UTC Timestamp for the END of the event (game end for example) """
        self.event_type: ALEventType = ALEventType(value=event_dict.get('eventType'))
        """ One of various event types (see: ALEventType) """


class GameEvent(Event):
    """ Event sub class for 'game' events """
    def __init__(self, event_dict: dict):
        super().__init__(event_dict)
        self.xp_progress: int = event_dict.get('xpProgress')
        """ XP progress to next level """
        self.game_length: int = event_dict.get('gameLength')
        """ game length (in minutes) """
        self.legend_played: str = event_dict.get('legendPlayed')
        """ Which legend was played in this game """
        self.rank_score_change: str = event_dict.get('rankScoreChange')
        """ Rank Score Change (if any) """
        self.game_data_trackers: list[DataTracker] = list()
        """ List of DataTrackers that had values in this game """
        tracker: dict
        for tracker in event_dict.get('event'):
            self.game_data_trackers.append(DataTracker(tracker))


class SessionEvent(Event):
    """ event subclass for "session" events (leave / join) """
    def __init__(self, event_dict: dict):
        super().__init__(event_dict)
        event_detail: dict = event_dict.get('event')
        self.action: str = event_detail.get('action')
        """ string of action 'leave' 'join' """
        if self.action == 'leave':
            self.session_duration: int = event_detail.get('sessionDuration')
        else:
            self.session_duration: int = 0


class LevelEvent(Event):
    """ event subclass for 'level' events (level up) """
    def __init__(self, event_dict: dict):
        super().__init__(event_dict)
        self.new_level = event_dict.get('event').get('newLevel')
        """ The new level achieved """


class RankEvent(Event):
    """ event subclass for 'rank' events (ranked progression)"""
    def __init__(self, event_dict: dict):
        super().__init__(event_dict)
        self.new_rank: str = event_dict.get('event').get('newRank')
        """ The new rank achieved """


def event_factory(event_dict: dict) -> Event:
    """ a factory method for the different event types"""
    event_classes = {
        "Session": SessionEvent,
        "Level": LevelEvent,
        "Game": GameEvent,
        "Rank": RankEvent
    }
    return event_classes[event_dict.get('eventType')](event_dict)


class GlobalInfo:
    """ a data structure for the global player info """
    # pylint: disable=too-many-instance-attributes
    # I'm fully aware, this is a dataclass, so pylint can just get over itself
    # Definition of local data structures
    class Bans:
        """ data structure for player bans """
        def __init__(self, bans_dict: dict):
            self.reason: str = bans_dict.get('last_banReason')
            """ Reason of last ban (still has value even when ban is over) """
            self.is_active: bool = bool(bans_dict.get('isActive'))
            """ True if ban is currently active """
            self.seconds_remaining: int = bans_dict.get('remainingSeconds')
            """ Remaining time (in seconds) of current ban """

    class Rank:
        """ data structure for player rank information """
        def __init__(self, rank_dict: dict):
            self.score: int = rank_dict.get('rankScore')
            """ Current rank score """
            self.tier: str = rank_dict.get('rankName')
            """ Rank Tier (Bronze, Gold, etc...) """
            self.division: int = rank_dict.get('rankDiv')
            """ Rank Division (I, II, etc...) """
            self.ladder_pos_platform: int = rank_dict.get('ladderPosPlatform')
            """ Ladder rank for Apex Predators, will show -1 for others """
            self.image_url: str = rank_dict.get('rankImg')
            """ Image icon for current rank """
            self.season: str = rank_dict.get('rankedSeason')
            """ season / split for current rank (ex: season08_split_2) """

        @property
        def rank_division_roman(self) -> str:
            """ return the rank Division Roman Numeral representation (I II III IV) """
            div: int = self.division
            if div == 1:
                return "I"
            if div == 2:
                return "II"
            if div == 3:
                return "III"
            if div == 4:
                return "IV"
            return ""

    def __init__(self, global_dict: dict):
        self.name: str = global_dict.get('name')
        """ Player Nick Name """
        self.uid: int = global_dict.get('uid')
        """ Player UID (guaranteed unique) """
        self.avatar: str = global_dict.get('avatar')
        """ Player Origin /EA Avatar """
        self.platform: ALPlatform = ALPlatform(value=global_dict.get('platform'))
        """ Platform (PC, XBOX, etc...) """
        self.level: int = global_dict.get('level')
        """ Current level of the player """
        self.to_next_level_percent: int = global_dict.get('toNextLevelPercent')
        """ Percent progress to next level (1-100) """
        self.bans: GlobalInfo.Bans = GlobalInfo.Bans(bans_dict=global_dict.get('bans'))
        """ Active / latest ban data """
        self.rank: GlobalInfo.Rank = GlobalInfo.Rank(rank_dict=global_dict.get('rank'))
        """ Current rank detail """
        self.badges: list[GameInfo.Badge] = list()
        if global_dict.get('badges'):
            for badge in global_dict.get('badges'):
                self.badges.append(GameInfo.Badge(badge_dict=badge))


class RealtimeInfo:
    """ a data structure for the player's real time information """
    def __init__(self, realtime_dict: dict):
        self.lobby_state: str = realtime_dict.get('lobbyState')
        """ open if lobby is open """
        self.is_online: bool = bool(realtime_dict.get('isOnline'))
        """ True if player is currently online """
        self.is_in_game: bool = bool(realtime_dict.get('isInGame'))
        """ True if player is currently in a game """
        self.can_join: bool = bool(realtime_dict.get('canJoin'))
        """ True if player can join party """
        self.party_full: bool = bool(realtime_dict.get('partyFull'))
        """ True if the player's party is full """
        self.selected_legend: str = realtime_dict.get('selectedLegend')
        """ Currently selected legend in the lobby """


class DataTracker:
    """
    Data structure for badges

    Note:
        if the tracker's rank is not available, and empty dictionary will be used
    """
    class TrackerRank:
        """ Data structure for the rank for the stat being tracked """
        def __init__(self, tracker_rank_dict: dict):
            self.position: int = tracker_rank_dict.get('rankPos')
            """ Position of rank """
            self.percent: float = float(tracker_rank_dict.get('topPercent'))
            """ Percentile of rank (lower number is better) """

    def __init__(self, data_trackers_dict: dict):
        self.name: str = data_trackers_dict.get('name')
        """ Descriptive name of the tracker """
        self.value: int = data_trackers_dict.get('value')
        """ Numerical value of the tracker """
        self.key: str = data_trackers_dict.get('key')
        """ Unique 'key' for the tracker """
        self.category: str = self.key.lstrip('specialEvent_')
        """
        Aggregate key for combining 'specialEvent' and 'season' data with regular data

        Note:
            Special Event data is the same as regular but has a different key for
            statistical purposes it makes sense to just drop the 'specialEvent' prefix
            or any 'season' suffix
        """
        if self.category.startswith('wins_'):
            self.category = 'wins'
        if self.category.startswith('kills_'):
            self.category = 'kills'
        rank_dict: dict = data_trackers_dict.get('rank')
        if not rank_dict:
            rank_dict = {'rankPos': -1, 'topPercent': -1.0}
        else:
            rank_dict = data_trackers_dict.get('rank')
        self.tracker_rank = DataTracker.TrackerRank(tracker_rank_dict=rank_dict)
        """
        Rank of the stat for this legend

        Note:
            * Values are -1 if they were not part of the API query
        """


class ImgAsset:
    """ data structure for image assets """
    def __init__(self, image_asset_dict: dict):
        self.icon: str = image_asset_dict.get('icon')
        """ URL to the icon of the image """
        self.banner: str = image_asset_dict.get('banner')
        """ URL to the banner of the image """


class Legend:
    """ data structure for a player's legend legend """
    def __init__(self, legend_name: str, legend_dict: dict):
        self.name: str = legend_name
        """ Legend name """
        self.data_trackers: list[DataTracker] = list()
        """ list of all known data trackers for the legend """
        if legend_dict.get('data'):
            data_tracker: dict
            for data_tracker in legend_dict.get('data'):
                self.data_trackers.append(DataTracker(data_tracker))
        self.img_assets: ImgAsset = ImgAsset(legend_dict.get('ImgAssets'))
        """ urls to icons / banner images"""
        if legend_dict.get('gameInfo'):
            self.game_info: GameInfo = GameInfo(legend_dict.get('gameInfo'))
            """ GameInfo for the legend """


class ALPlayer:
    """
    This class encapsulates the player and associated data retrieved from the API calls

    Discussion:
        The player class is intended as an encapsulated representation of the data from the api.
    """

    def __init__(self, basic_player_stats_data: dict, events: list = None):
        """ Init the object with the player's name and platform """
        self.global_info: GlobalInfo = GlobalInfo(global_dict=basic_player_stats_data['global'])
        """ Contains all the global info for the player """
        self.realtime_info: RealtimeInfo = RealtimeInfo(
            realtime_dict=basic_player_stats_data['realtime']
        )
        """ Contains the RealtimeInfo for the player """
        self.timestamp_last_checked: int = arrow.utcnow().int_timestamp
        """ Contains the timestamp (UTC) that the player was created / data loaded """
        self.selected_legend: Legend = Legend(
            legend_name=basic_player_stats_data['legends']['selected']['LegendName'],
            legend_dict=basic_player_stats_data['legends']['selected']
        )
        """ Currently Selected Legend """
        self.all_legends: List[Legend] = list()
        """ List of all legends (and their stats) """
        for legend_name, legend_dict in basic_player_stats_data['legends']['all'].items():
            self.all_legends.append(Legend(legend_name=legend_name, legend_dict=legend_dict))
        self.events: List[Event] = list()
        """ List of all matches / events"""
        if events:
            event: dict
            for event in events:
                event_result = event_factory(event_dict=event)
                self.events.append(event_result)



""" Apex Legends API Python Module """
from .al_api import ApexLegendsAPI
from .al_base import ALPlatform, ALAction, ALHTTPExceptionFromResponse
from .al_domain import ALPlayer

__all__ = [
    'ApexLegendsAPI',
    'ALPlatform',
    'ALAction',
    'ALPlayer',
    'ALHTTPExceptionFromResponse'
]



"""
al_base.py
contains some of the base / utility classes and Enums
"""
from enum import Enum
import requests


def print_description(___class, indent=0, hide_values=False):
    """ prints the schema for the current object """
    print(' ' * indent + type(___class).__name__ + ':')
    indent += 4
    for k, value in ___class.__dict__.items():
        if not isinstance(value, list):
            v_list = [value]
        else:
            v_list = value
        for val in v_list:
            if '__dict__' in dir(val):
                print_description(val, indent)
            else:
                if hide_values:
                    print(' ' * indent + k)
                else:
                    print(' ' * indent + k + ': ' + str(val))


class ALPlatform(Enum):
    """
    Three platforms available
    - XBOX
    - PSN
    - PC
    """
    XBOX = "X1"
    """ Xbox """
    PSN = "PS4"
    """ Playstation (any) """
    PC = "PC"
    """ PC """


class ALAction(Enum):
    """
    Three actions available
    - INFO Return the players you're currently tracking
    - GET Return ALL tracked events for the player
    - ADD Adds the player for history collection
    - DELETE Removes the given user from the tracked users list
    """
    INFO = "info"
    """ Return the players you're currently tracking """
    GET = "get"
    """ Return ALL tracked events for the player """
    ADD = "add"
    """ Adds the player for history collection """
    DELETE = "delete"
    """ Removes the given user from the tracked users list """


class ALEventType(Enum):
    """
    The four different event types
      - SESSION
      - GAME
      - LEVEL
      - RANK
    """
    SESSION = 'Session'
    """ Session event (leave, join) """
    GAME = 'Game'
    """ Game event """
    LEVEL = 'Level'
    """ Level Up event """
    RANK = 'Rank'
    """ Rank change event """


class ALHTTPExceptionFromResponse(Exception):
    """ Exception raised for errors in the http request. """
    def __init__(self, response: requests.Response):
        self.message = f'Return Code: {response.status_code} - {response.text}'
        super().__init__(self.message)



""" setup script for apex_legends_api """
from setuptools import setup

with open('README.md') as f:
    long_description = f.read()

with open('requirements.txt') as f:
    requirements = f.readlines()

setup(
    name='apex-legends-api',
    version='1.5.0',
    packages=['apex_legends_api'],
    python_requires='>=3.6.*',
    url='https://github.com/johnsturgeon/apex-legends-api',
    license='MIT',
    author='John Sturgeon',
    author_email='john.sturgeon@me.com',
    install_requires=requirements,
    description='Python wrapper for https://apexlegendsapi.com',
    long_description=long_description,
    long_description_content_type='text/markdown'
)
